from .textools import style_text, Align
from . import logictools  
import os

def spacer(times):
    for _ in range(times): print()

def underline(line: str="—", hue: str="", alone=False, return_str: bool = False):
    term_width = logictools.get_term_size(True)
    output = ""
    line = style_text(line*term_width, hue)
    if alone:
        if not return_str: print()
        output += "\n"
    if not return_str: print(line)
    else: output += line
    if alone:
        if not return_str: print()
        output += "\n"
    return output

def clear(header=None):
    os.system('cls' if os.name == 'nt' else 'clear')
    if header: print(header)


from typing import Any, NoReturn, Callable
import unicodedata
import random
import shutil
import re

def copy(data:Any, times:int = 2) -> list[Any]|NoReturn:
    def gave_up() -> NoReturn:
        err = "Invalid input! <times> got " \
            + f"{times!r} ({type(times)}). Could not " \
            + "convert to a whole number"
        err = _wrap_text(_color(err, "red"), inline=True,
              list_order="TypeError", indent=11)
        raise TypeError(err)
    
    try: return [data for _ in range(int(times))]
    except ValueError: gave_up()
    
def make_progress_bar(pct, width=20):
    filled = int(pct * width)
    return "█" * filled + "▒" * (width - filled)

def rated_bar(pct, width=20):
    control = int(pct * width)
    filled = min(int(pct * width), 7)
    first = "█" * filled + "▒" * (7 - filled)
    rate = format_order(f"{(pct * 100):.2f}")
    last = "▒" * 7
    if pct > 0.6:
        pct -= 0.6
        filled = min(int(pct * width), 7)
        last = "█" * filled + "▒" * (7 - filled)
     
    return f"{first}{rate}%{last}"

def percent_colored(progress: int|float, head:str = "",
                    tail:str = "%") -> str:
    sign = f"+{head}" if progress >= 0 else f"-{head}"
    hue  = "green" if progress >= 0 else "red"
    prog = abs(progress)
    return color(f"{sign}{prog:.2f}{tail}", fg=hue)

def not_any(*args) -> bool:
    if len(args) == 1:
        args = args[0]
        if not isinstance(args, (list, tuple)):
            return not bool(args)
            
    for arg in args: 
        if not arg: return True
    return False 

def not_all(*args) -> bool:
    if len(args) == 1: return not_any(*args)
            
    for arg in args: 
        if arg: return False
    return True

def any_in(*args, eq=None) -> bool:
    if len(args) == 1:
        args = args[0]
        if not isinstance(args, (list, tuple)):
            return args in eq
            
    for arg in args: 
        if arg in eq: return True
    return False 

def any_eq(*args, eq=None) -> bool:
    if len(args) == 1:
        args = args[0]
        if not isinstance(args, (list, tuple)):
            return args == eq
            
    for arg in args: 
        if arg == eq: return True
    return False

def any_is(*args, eq=None) -> bool:
    if len(args) == 1:
        args = args[0]
        if not isinstance(args, (list, tuple)):
            return args is eq
            
    for arg in args: 
        if arg is eq: return True
    return False

def any_rel(args: list, rel: Any, func: Callable) -> bool:
    for arg in args:
        if func(arg, rel): return True
    return False

def all_in(*args, eq=None) -> bool:
    if len(args) == 1: return any_in(*args)
            
    for arg in args:
        if arg not in eq: return False
    return True

def all_eq(*args, eq=None) -> bool:
    if len(args) == 1: return any_eq(*args)
            
    for arg in args:
        if arg != eq: return False
    return True

def all_is(*args, eq=None) -> bool:
    if len(args) == 1: return any_is(*args)
            
    for arg in args:
        if arg is not eq: return False
    return True

def all_rel(args: list, rel: Any, func: Callable) -> bool:
    for arg in args:
        if not func(arg, rel): return False
    return True

def shave(num: int | float, limit: int | float) -> tuple:
    major = int(num / limit)
    shaved = num - major * limit
    if shaved < 0: 
        # Fixes weird negative remainders for huge 
        # units (e.g., trillionenniums) — magic number
        shaved = abs(shaved) / 19 # DON'T TOUCH!
    if  shaved >= limit:
        shaved -= int(shaved / limit) * limit 
        major  += int(shaved / limit)
        
    return round(shaved), major

def get_term_size(width: bool = False) -> int:
    if not width:
        return shutil.get_terminal_size().columns or 80
    return shutil.get_terminal_size((80, 20)).columns or 80

def visual_width(s: str) -> int:
    clean = strip_ansi(s)
    width = 0
    for ch in clean:
        width += 2 if unicodedata.east_asian_width(
            ch) in ['F', 'W'] else 1
    return width

def number_padding(num, pad=3):
    return str(num).rjust(pad)

def format_order(order: str, deno=2, form="0"):
    if not isinstance(form, str): 
        raise TypeError("form should be a string")
    if not isinstance(deno, int) or deno < 0:
        raise TypeError("deno should be an integer")
    
    try: length = len(str(int(float(order))))
    except ValueError: length = len(str(order))
    if length < deno:
        fill = (deno - length) * form
        order = f"{fill}{order}"
    return str(order)

def variance(new: int|float, old: int|float)->int|float:
    for arg in [new, old]:
        if not isinstance(arg, (int, float)):
            raise TypeError(f"{arg} is supposed to"
                            +"be int or float")
    
    if old: return ((new - old) / old) * 100
    elif not new and not old: return 0
    else: return 100

def strip_ansi(s: str) -> str:
    return re.sub(r'\x1B[@-_][0-?]*[ -/]*[@-~]','',s)


# Private helpers
def _color(text, fg: str|None = None, bg: str|None = None, 
          bold:bool = False, underline:bool = False)-> str:
    colors = {
        "black": 30, "red": 31, "green": 32, 
        "yellow": 33, "blue": 34, "magenta": 35,
        "cyan": 36, "white": 37,
        "gray": 90, "lightred": 91, "lightgreen": 92,
        "lightyellow": 93, "purple": 94,
        "lightmagenta": 95, "lightcyan": 96
    }

    styles = []
    
    if bold: styles.append("1")
    if underline: styles.append("4")
    if fg in colors: styles.append(str(colors[fg]))
    if bg in colors: styles.append(str(colors[bg]+10))

    if styles: 
        return f"\033[{';'.join(styles)}m{text}\033[0m"
    return text

def _wrap_text(text:str, indent:int = 0, pad:int = 0, 
              inline:bool = False, list_order:str = ''
              ) -> str:
    width        = get_term_size()
    styled_words = text.split()
    
    length   = len(list_order) if inline else 0
    trailing = length / (10 ** len(str(length))
               ) if not inline else 0
    edge = 1 + trailing
    if length > 9 and not inline:
        trailing = length * 2 / 10
    margin = 1 + trailing if length > 9 else edge
    
    line_len = len(list_order)
    result   = list_order if not inline else ""
    if pad:
        result   = " " * (pad - 1)
        line_len = pad
    for i, word in enumerate(styled_words):
        used = line_len + visual_width(word)
        if used + margin > width:
            result += '\n' + ' ' * indent + word
            line_len = indent + visual_width(word)
        else:
            result += (' ' if result else '') + word
            line_len += visual_width(
                word) + margin / (2 if margin >= 2 
                else 1)
    
    return result



"""
Module for all list-related tools
"""
from .logictools import format_order, visual_width, any_is
from .exceptions import ListError, ExitError, warning
from collections.abc import KeysView as keys
from .textools import style_text, wrap_text
from collections.abc import Iterable
from typing import Any, Callable
from . import console

def choose(options:dict, cursor:str = ">>> ",
          hue:str = "magenta", getch:bool = False,
          src: Callable = None, proxy:bool = False,
          back:bool = True):
    """
Custom input menu for choosing an option from a list

Usage:
  - This function does not list the options.
  - The values of options (dict) should be functions. If
    not, use pick_from(...) instead.
  - Write 'clear' to reset the source screen.
  - This function works best when paired with the 
    function list_items(...)
  - For a real-world usage example of this function,
    check out my CLI app Habitrax on GitHub:
    https://github.com/2kDarki/Habitrax

Args:
    options (dict): A dictionary with functions as values
    cursor   (str): The prompt cursor
    hue      (str): Color of the prompt cursor
    getch   (bool): If true returns the index position of
                    the chosen function rather than calling
                    it. NB: if needed, you can use this
                    flag for options that are lists or
                    non-callable dictionary values to 
                    return their index position then use
                    pick_from(...) to get the value
    src (callable): The name of the function/method where
                    this function was called from. Needed
                    for clearing the screen and recalling
                    where you were
    proxy   (bool): Needed in rare cases where the
                    function/method that called this
                    function was called by another
                    function/method and it needs to know
                    that. See Habitrax for example
    """
    if back:
        print(f"{format_order(len(options)+1, form=' ')}."
              +" Back")
    while True:
        try:
            choice = input(style_text(cursor, hue))
            choice = int(choice) - 1
            if 0 <= choice < len(options):
                if getch: return choice
                elif proxy: pick_from(options, choice)(1)
                else: pick_from(options, choice)()
                break
        except BaseException as e:
            if isinstance(e, ExitError()): 
                if isinstance(e, EOFError): print()
                print(end="\n")
                console.underline(hue=hue)
                return                
            if choice == "clear" and src:
                if proxy:
                    console.clear()
                    src(1)
                    return
                console.clear()
                src()
                return
        if choice == len(options):
            if src:
                print()
                console.underline(hue=hue)
            return
        warning("Choose a number between 1 and "
              +f"{len(options)+1}")

def pick_from(storage: list|tuple|dict, choice:int,
              index: int | None = None) -> Any:
    """
Takes a dictionary/list/tuple/dict_keys and a numeric index

Args:
    storage: A dictionary, list, tuple, or dict keys
    choice: Index position of value needed
    index: If None (default), raises StopIteration when the 
           provided choice is greater than storage's 
           index positions. Otherwise use index as fallback 
           choice
           

Returns:
    - corresponding value
    - rather than throwing
    """
    if choice < 0: choice = len(storage) - abs(choice)
    if isinstance(storage, (list, tuple, keys)):
        return list(storage)[choice]
    return next((storage[opt] for pos, opt in
           enumerate(storage) if pos == int(choice)))

def format_items(items:list) -> list:
    """Format items to give out true name"""
    return [item.__name__ if callable(item) else item for
            item in items]

def list_items(items:list, guide: str|None = None,
               style: str|None = "number",
               spaced:bool = False) -> None:
    """
Displays all items in <items> as an ordered list

Args:
    items: A list, tuple or dictionary
    guide: Optional header for the items
    spaced: Adds empty lines between items
    style: Style of the ordered list:
           number, roman, alpha, and bullet
    """
    items = format_items(items)
    deno, indent, sfx, order = 0, 0, "", ""
    styles = ["number", "roman", "alpha", "bullet"]
    if style and style not in styles:
        err(style, f"list order style {tuple(styles)}")
    elif style:
        if styles.index(style) < 3: sfx = "."
        style  = ListOrder(style, suffix=sfx)
        deno   = max((visual_width(style(i+1))-1 for i, x in
                 enumerate(items)))
        deno   = max(2, deno)
        indent = deno + 2

    if guide:
        print(style_text(guide, underline=True)+":")
    for opt, item in enumerate(items):
        if style: order = format_order(style(opt),deno," ")
        print(wrap_text(item, indent, order=order))
        if spaced: print()

def flatten(data:Iterable, to:type = list) -> list|tuple:
    """
Recursively flattens any iterable (except str/bytes) into a flat list or tuple.

Args:
    data (Iterable): Input iterable, possibly nested.
    to (list or tuple): What to return the flattened
                        data as

Returns:
    Flattened version of the iterable.
    data: If input is not an iterable.
    """
    
    if not any_is(list, tuple, eq=to):
        err(to, "function (list or tuple)")
    if not isinstance(data, Iterable) or isinstance(
        data, (str, bytes)): return data

    flat = []
    for item in data:
        if isinstance(item, Iterable) and not (isinstance(
           item, (str, bytes))): flat.extend(flatten(item))
        else: flat.append(item)

    return to(flat)

def normalize(array:list, use:Callable|None = None)-> list:
    normalized = []
    
    for obj in array:
        el = obj
        if use is not None: el = use(obj)
        if el not in normalized: normalized.append(el)
    
    return normalized

class ListOrder:
    """Class for turning an index into a list style"""
    def __init__(self, style:str = "number", prefix:str="",
                 suffix:str = ".", start:int = 1):
        self.style  = style
        self.prefix = prefix
        self.suffix = suffix
        self.start  = start
        self.styles = {
            "number": self._number,
            "roman": self._roman,
            "alpha": self._alpha,
            "bullet": self._bullet
        }

    def format(self, index:int) -> str:
        """Format index to list style"""
        base = self.styles.get(self.style, self._number)(
               index + self.start)
        return f"{self.prefix}{base}{self.suffix}"

    @staticmethod
    def _number(n:int) -> str: return str(n)

    @staticmethod
    def _roman(n: int) -> str:
        """Turn a number into a roman numeral"""
        romans = [
            (1000,'M'), (900,'CM'), (500,'D'), (400,'CD'),
            (100,'C'), (90,'XC'), (50,'L'), (40,'XL'),
            (10,'X'), (9,'IX'), (5,'V'), (4,'IV'), (1,'I')
        ]
        result = ""
        for val, sym in romans:
            while n >= val:
                result += sym
                n -= val
        return result

    @staticmethod
    def _alpha(n: int) -> str:
        """Turn a number into an alpha numeral"""
        # 1 -> A, 2 -> B, ..., 26 -> Z, 27 -> AA
        result = ""
        n -= 1
        while n >= 0:
            result = chr(n % 26 + 65) + result
            n = n // 26 - 1
        return result

    @staticmethod
    def _bullet(_) -> str:  return "•"

    def __call__(self, index):
        return self.format(index)

def err(cause, required):
    """Helper for raising list-related errors"""
    raise ListError(cause=cause, required=required)



from .__const__ import EIGHTEENTH, UNITTHRESH
from .textools import pluralize, style_text
from .logictools import shave, format_order
from .exceptions import warning, TimeError
from typing import NoReturn, Any
from .textools import wrap_text
from datetime import datetime
from .console import clear
from math import inf

def format_time(time: int|float, set_to:str = "second", 
                hue:str|None = None, tense:str|None = None, 
                precise:bool = True, faulty:bool = True,
                only=False) -> str:
    """
Returns a fuzzy human-readable time string

Args:
       time: a positive or negative number.
             If negative, it gives out a string like
             "24 minutes and 30 seconds ago". Else it
             returns "24 minutes and 30 seconds"
     set_to: specify what order the argument 'time' 
             is in. <time=1, set_to='hour'> is greater
             than <time=56, set_to='second'>.
             Default=second
        hue: the color of the returned string
      tense: temporal reference ('future' or 'past').
             if 'future': returns "in 2 minutes"
             if 'past': returns "2 minutes ago"
             if None: returns "2 minutes"
    precise: whether or not to return the time string
             down to the second. E.g., if precise=True
             it returns "55 years, 9 months, 2 weeks, 
             3 days, 14 hours, 37 minutes and 41 
             seconds". Else returns "55 years and 9 
             months"
     faulty: used when the time is greater than a month.
             The conversion used are faulty by default
             (1 month ≠ 4 weeks) so it is used for 
             corrections. You can leave it as-is (it 
             just takes a few milliseconds longer to
             format the time)
    """
    if set_to not in UNITTHRESH:
        err(set_to, "period of time")
 
    try: past = time < 0
    except TypeError: err(time, set_to)
    
    if tense and tense not in ["future", "past"]:
        err(tense, 
            "temporal reference ('future' or 'past')")
    
    if past: 
        time  = abs(time)
        tense = "past"
    
    units = list(UNITTHRESH.keys())     
    for unit in units[:units.index(set_to)]:
        time *= UNITTHRESH[unit]

    if faulty: time = get_time_lapsed(time, num=True)
    if not precise and time > 60:
        return style_text(time_lapsed(time, tense=tense), 
               hue)
        
    def get_timeunits(time) -> list[list[int|float, str]]:
        timeunits = []
        limits    = [UNITTHRESH[u] for u in units]

        for limit, unit in zip(limits, units):
            if only:
                if unit == set_to:
                    timeunits.append([time, unit])
                    if len(timeunits) > 1:
                        timeunits = timeunits[1:]
                    break
            if time >= limit:
                time = round(time)
                shaved, time = shave(time, limit)
                timeunits.append([shaved, unit])
            else:
                timeunits.append([time, unit])
                break

        return timeunits
        
    timeunits = get_timeunits(time)
    for i, timeunit in enumerate(timeunits):
        time, unit = timeunit
        unit = pluralize(time, unit)
        timeunits[i] = [time, unit]
    
    return style_text(format_time_string(timeunits[::-1], 
           tense=tense), fg=hue)

def convert(time:str, to:str = "hours") -> int | float:
    """
Takes a 24-hour notation time string (e.g., "23:45")
Converts the time string to the specified unit:
    if to = hours: returns 23.75
    if to = minutes: returns 1425
    if to = seconds: returns 85500

The time string should be in one of these 3 formats:
    1. hh
    2. hh:mm
    3. hh:mm:ss
    """   
    units, limits = get_unit_limits(plural=True)
    try:
        parts = [int(part) for part in time.split(":")][:3]
        for part, limit, unit in zip(parts, limits[::-1], 
                                             units[::-1]):
            if part not in range(limit):
                xrange = "is below minimum" if (part 
                       < 0) else "exceeds maximum"
                raise TimeError(f"{part!r} {xrange}"
                      +f" value allowed for {unit}")
    except (AttributeError, ValueError):
        err(time, "time string (hh, hh:mm or hh:mm:ss)")
    
    length = 0        
    for i, part in enumerate(parts):
        length += part / 60**i
    
    for pos, unit in enumerate(units[::-1]):
        if to == unit: return length * (60**pos)
    
    err(to, "unit (hours, minutes or seconds)")

def timestamp(iso_str: str | datetime | None = None,
              date_only: bool = False, short: bool = False,
              mini: bool = False) -> str:
    """
Converts ISO timestamp into a human-readable string.

Args:
    iso_str: ISO 8601 datetime string or object, or None.
    date_only (bool): If True, omits time (returns date only).
    short (bool): If True, returns compact date (e.g., "14 Sep 2025").
    mini (bool): If True, returns compact date (e.g., "2025-09-14")

Returns:
    str: Formatted timestamp string.
    """
    if iso_str is None: iso_str = datetime.now().isoformat()
    try: iso_str = to_iso(iso_str)
    except TimeError: pass
    
    try: dt_obj = datetime.fromisoformat(iso_str)
    except (ValueError, TypeError):
        err(iso_str, 
           "iso string (YYYY-MM-DD) or datetime object")
                        
    if mini: return dt_obj.strftime("%Y-%m-%d")
    elif short: return dt_obj.strftime("%d %b %Y")
    elif date_only: return dt_obj.strftime("%A, %d %B %Y")
    return dt_obj.strftime("%A, %d %B %Y • %H:%M")

def time_lapsed(last:int|float, now:int|float|None = None, 
                tense:str|None = "past") -> str:
    """
Returns a fuzzy human-readable time difference
string (e.g., "3 hours and 12 minutes ago").
    """
    if tense and tense not in ["future", "past"]:
        err(tense, "temporal reference")
    
    def get_lapsed(last:int|float, now:int|float|None=None
                  ) -> tuple[str, int|float]:
        lapsed = last
        if now: lapsed = now - last
        s = pluralize(lapsed, "second")
        string = f"{int(lapsed)} {s} ago"
        return string, lapsed

    s, lapsed = get_lapsed(last, now) if now \
           else get_lapsed(last)
 
    # Checks if user has tempered with device date
    tempered = lapsed < -5
  
    if -5 <= lapsed < 1: return "Just now"
  
    # Thresholds for escalation to next unit
    units, limits = get_unit_limits("year")
    prv_unit      = "second"
    for unit, limit in zip(units[1:], limits[:-1]):
        if lapsed >= limit:
            s, lapsed = format_unit(lapsed, limit, unit,
                        prv_unit, tense)
            prv_unit  = unit
        else: break

    return s if not tempered else warning(
        'Tempered with device date', inline=True)

def format_time_string(*args, tense:str|None=None) -> str:
    def tensed(body:str) -> str:
        head = "in "  if tense == "future" else ""
        tail = " ago" if tense ==   "past" else ""
        return f"{head}{body}{tail}"

    if len(args) == 1 and isinstance(args[0], list):
        args = args[0]

    parts = []
    try:
        parts = [f"{unit[0]} {unit[1]}" for unit in args if
                    unit[0] != 0]
    except TypeError: 
        if args[0] != 0: parts = [f"{args[0]} {args[1]}"]
    
    if len(parts) > 1: return tensed(', '.join(parts[:-1])
                            + f" and {parts[-1]}")
    elif len(parts) == 1: return tensed(parts[0])
    else: return "just now"

def format_unit(lapsed:int|float, unit:int,
                major_label:str, minor_label:str, 
                tense:str|None = None) -> tuple[str, int]:
    major = int(lapsed / unit)
    minor = round(lapsed - major * unit)
    major_str = pluralize(major, major_label)
    minor_str = pluralize(minor, minor_label)
    return format_time_string([major, major_str],[minor, 
           minor_str], tense=tense), major
        
def get_time_lapsed(iso_str:str, fuzzy:bool = True, 
                    num:bool = False, 
                    _fixed:str|None=None) -> str|int|float:
    try: before = datetime.fromisoformat(iso_str)
    except (TypeError, ValueError):
        if num: pass
        else: err(iso_str, "iso string (YYYY-MM-DD)")
              
    now   = EIGHTEENTH if _fixed else datetime.now()
    years = 0
    if not num:
        diff   = now - before
        years  = int(diff.days / 365)
        lapsed = diff.total_seconds()
    else:
        years  = int(iso_str / 60 / 60 / 24 / 365)
        lapsed = iso_str
        fuzzy  = False
    if years: lapsed -= lapsed - 29030400 * years
    return time_lapsed(lapsed) if fuzzy else lapsed

def to_iso(iso_str:str) -> str:
    if isinstance(iso_str, datetime): 
        return iso_str.isoformat()
    try:
        timestamp = datetime.fromisoformat(iso_str)
        return timestamp.isoformat()
    except (ValueError, TypeError):
        err(iso_str, "iso string (YYYY-MM-DD)")

def get_age(b_day:str, years:bool = True, 
            _fixed:bool = False) -> int|str:
    try: b_day = datetime.fromisoformat(b_day)
    except (ValueError, TypeError):
        err(b_day, "iso string (YYYY-MM-DD)")
    
    today = EIGHTEENTH if _fixed else datetime.now()
    if years: return int((today - b_day).days / 365)
    lapsed = get_time_lapsed(to_iso(b_day),0,_fixed=_fixed)
    return format_time(lapsed, "second", faulty=False)
    
def get_unit_limits(upto:str = "hour", plural:bool = False
                   ) -> tuple[str, int]:
    if upto not in UNITTHRESH.keys():
        err(upto, "unit of time")
    
    units, limits = [], []
    for unit, limit in UNITTHRESH.items():
        u = pluralize(0, unit) if plural else unit
        units.append(u)
        limits.append(limit)
        if unit == upto: return units, limits

def clock(time: str|int|float, notation:str = "24", 
          meridiem: str|None = None) -> str:
    try: time = float(time)
    except ValueError: err(time, "number") 
    
    if notation != "24":
        if notation == "12":
            am = ["am", "a.m", "a.m."]
            pm = ["pm", "p.m", "p.m."]
            if meridiem and meridiem in am + pm:
                if meridiem in pm: time += 12
            else: err(meridiem, "meridiem (a.m or p.m)")
        else: err(notation,"hour notation (12 or 24)")
        
    hours   = int(time)
    minutes = float(round((time - hours) * 60, 7))
    seconds = int(round((minutes - int(minutes)) * 60, 7))
    
    times = [hours, minutes, seconds]
    units, limits = get_unit_limits(plural=1)
    unitlimits = list(zip(units, limits))[::-1]
    for time, (unit, limit) in zip(times, unitlimits):
        if int(time) not in range(limit):
            xrange = "is below minimum" if time < 0 \
                else "exceeds maximum"            
            raise TimeError(f"{time!r} {xrange} value "
                          + f"allowed for {unit}")
    
    hrs  = format_order(hours)
    mins = format_order(int(minutes))
    sec  = format_order(seconds)
    
    if seconds: return f"{hrs}:{mins}:{sec}"
    return f"{hrs}:{mins}"

def err(cause:Any, required:str) -> NoReturn:
    raise TimeError(cause=cause, required=required)



from datetime import datetime
from math import inf

EIGHTEENTH = datetime(2025, 3, 23)
UNITTHRESH = {
            "second": 60, 
            "minute": 60, 
              "hour": 24, 
               "day": 7, 
              "week": 4, 
             "month": 12,
              "year": 100,
           "century": 10,
        "millennium": 1000,
     "millionennium": 1000,
     "billionennium": 1000,
    "trillionennium": inf
}




from typing import Any, NoReturn
import unicodedata
import textwrap
import time
import re
import os

from .exceptions import validate, InputError
from . import logictools


class Align:
    def __init__(self, offset: int = 0):
        self.offset = offset

    def right(self, arg:str) -> str:
        try: return wrap_text(arg.strip())
        except AttributeError: err(arg, "string")
    
    def center(self, arg:Any, line:str = " ", 
               hue:str|None = None, 
               line_hue:str|None = None, pad:int = 0, 
               get_pad: bool = False) -> str | tuple:
        # Validate parameters
        validate([line, str, "string"],
                 [pad, int, "natural number", "less", 0],
                 err=err)
        
        arg        = str(arg)
        term_width = logictools.get_term_size(1)-self.offset
        vis_width  = logictools.visual_width(arg)
        total_pad  = max(term_width - vis_width, 0)
        left_pad   = total_pad // 2
        right_pad  = total_pad - left_pad
    
        if get_pad:
            return left_pad, total_pad, right_pad

        left   = style_text(line *  left_pad, line_hue)
        right  = style_text(line * right_pad, line_hue)
        middle = style_text(arg, hue)
        
        if len(arg) > term_width:
            if pad: total_pad += pad * 2
            return wrap_text(middle, from_center=[hue,
                   line, line_hue, total_pad]) 
        
        return f"{left}{middle}{right}"
    
    def left(self, arg, pad: int = 4) -> str:
        validate([pad, int,  "natural number", "less", 0],
                 err=err)
        return wrap_text(str(arg), pad, pad)


def keep_split(text:str, sep=" "):
    parts = text.split(sep)
    if len(parts) < 2: return parts
    elif len(parts) == 2:
        if parts[1] == "": return [parts[0]+sep]
        else: return [parts[0]+sep, parts[1]]
    if parts[-1] == "":
        parts[-2] += sep
        parts = parts[:-1]
    
    final = []
    for i, part in enumerate(parts):
        if not part.endswith(sep):
            if i != len(parts) - 1: part += sep
        final.append(part)
    return final


def transmit(*msg: str, speed: int|float = 0.1,
             hold: int|float = 0.25, 
             hue: str|None = None,
             inline: bool = False, 
             inlined: list | None = None) -> None|str:
    validate([[speed, hold], [int, float], 
             "positive number", "less", 0], err=err)

    if len(msg) == 1: msg = str(msg[0])
    else: msg = " ".join([str(p) for p in msg])
    paragraphs = msg.split("\n\n")
    text       = ""
    para_len   = len(paragraphs)
    for pi, paragraph in enumerate(paragraphs):
        lines = paragraph.split("\n")
        for li, line in enumerate(lines):
            styled_words = keep_split(line, "[0m")
            sentence     = ""
            for word in styled_words:
                styled_word = style_text(word, hue)
                sentence += styled_word
            if not inlined: sentence = wrap_text(sentence)
            else: sentence = wrap_text(sentence, *inlined)
            if inline: 
                text += sentence
                if li != len(lines) - 1: text += "\n"
                continue
            for ch in sentence:
                delay = speed
                if ch == " ": delay = hold
                if logictools.any_in(ch, 
                    eq=["[", "0", "m"]): delay = 0
                print(ch, end="", flush=True)
                time.sleep(delay)
            if len(lines) > 1 and li != len(lines) - 1: 
                print()
        
        if inline: 
            if pi != len(paragraphs) - 1: text += "\n\n"
        elif para_len > 1 and pi != para_len-1: print()
    
    if inline: return text
    if len(paragraphs) == 1: print()


def strip_ansi(s:str) -> str:
    if not isinstance(s, str): err(s, "string") 
    return re.sub(r'\x1B[@-_][0-?]*[ -/]*[@-~]', '', s)


def visual_width(s:Any) -> int:
    clean = strip_ansi(str(s))
    width = 0
    for ch in clean:
        width += 2 if unicodedata.east_asian_width(
                 ch) in ['F', 'W'] else 1
    return width

def pluralize(n: int|float, word:str) -> str:
    # Validate parameters
    validate([n, [int, float], "integer or float"],err=err)
    strip_ansi(word) # relegating validation
    
    if n == 1: return word
    else:
        if word.endswith('y') and len(word) > 3: 
            return word[:-1]+"ies"
        return word + 's'

def styled(text:str, get:bool = False) -> bool|list:
    def append(hues:list, array:list) -> list:
        for i, hue in enumerate(hues):
            if hue in codes: array.append(hues[i])
        return array
    
    validate([text, str, "string"], err=err)
    if not logictools.any_in("\033", "\x1b[", eq=text): 
        return False if not get else []
    if not get: return True
    if logictools.any_eq("\033", "\x1b[", eq=text): 
        return []
    
    chars = text.split()
    try: start = chars.index("[") + 1
    except ValueError: return []
    end   = chars.index("m")
    parts = [p for p in chars[start: end]]
    codes = "".join(parts).split(";")
    codes = [int(c) if len(c) == 2 else c for c in codes]
    
    bold      = "1"
    underline = "4"
    fgs       = [n for n in range(30, 38)]
    bgs       = [n for n in range(90, 97)]
    
    styles = []
    if bold in codes: styles.append(bold)
    if underline in codes: styles.append(underline)
    styles = append(fgs, styles)
    return   append(bgs, styles)

def style_text(text, fg: str = "", bg: str = "",
               underline: bool = False,
               bold: bool = False) -> str:
    def append(array:list, _type:str="fg") -> list:
        try:
            start, end = 30, 38
            if _type == "bg": start, end = 90, 97
            for n in range(start, end):
                if n in styles: array.append(str(n))
            return array
        except TypeError: return array
       
    validate([[underline, bold], bool, "boolean"], err=err)
    
    special = "\n" in str(text)
    clean  = strip_ansi(str(text))
    style  = []
    styles = styled(str(text), get=True) # for restyling
    if not styles: clean = str(text)
    colors = {
        "black": 30, "red": 31, "green": 32, 
        "yellow": 33, "blue": 34, "magenta": 35,
        "cyan": 36, "white": 37,
        "gray": 90, "lightred": 91, "lightgreen": 92,
        "lightyellow": 93, "purple": 94,
        "lightmagenta": 95, "lightcyan": 96
    }

    for c in [fg, bg]:
        if c and c not in colors: err(c, "color")
    
    if special:
        # the enhanced print function
        return transmit(str(text), hue=fg, inline=True)
    
    if bold or "1" in styles: style.append("1")
    if underline or "4" in styles: style.append("4")
    if fg: style.append(str(colors[fg]))
    else: style = append(style)
    if bg: style.append(str(colors[bg] + 10))
    else: style = append(style, "bg")
    
    if style:
        return f"\033[{';'.join(style)}m{clean}\033[0m"
    return clean

def iswhitespace(text:str) -> bool:
    for ch in text:
        if ch != " ": return False
    return True

def format_adj(text):
    last_ch   = ""
    formatted = ""
    disallowed = (" ", "‽")
    
    for ch in text:
        if iswhitespace(ch) and last_ch not in disallowed:
            ch = "‽"
        last_ch = ch
        formatted += ch
    
    return formatted

def flatten(text: str) -> str:
    return " ".join(text.split())

def wrap_text(text: str,  indent: int = 0,  pad: int = 0, 
              sub_indent: int = 0, inline: bool = False, 
              order: str = '', from_center: list = [],
              _internal: bool = False) -> str:
    validate([order, str, "string"],
        [[indent, pad], int, "natural number", "less", 0],
        [[inline, _internal], bool, "boolean"],
        [from_center, list, "list"], err=err)
    
    width = logictools.get_term_size()
    if not text: return ""
    if isinstance(text, str) and "\n" in text:
        args = [indent, pad, sub_indent, inline, order]
        return transmit(text, inline=True, inlined=args)
    if _internal:
        if isinstance(text, list):
            styled_words   = format_adj(text[0]).split("‽")
            rest           = text[1:]
        else:
            text         = str(text)
            styled_words = format_adj(text).split("‽")
    else:
        text         = str(text)
        styled_words = format_adj(text).split("‽")
        lines        = text.split("\n")
        if len(lines) > 1:
            _internal    = True
            styled_words = format_adj(lines[0]).split("‽")
            rest         = lines[1:]
    
    if from_center:
        h, l, lh, total_pad = from_center
        width -= total_pad
        centered, res = "", ""
        pos = 0
    
    length = visual_width(order)
    margin = 1
    
    line_len = visual_width(order)
    result   = order if not inline else ""
    
    if pad:
        if not inline: result = " " * pad
        if isinstance(text, str):
            text = " " * pad + text
        line_len = pad
    
    if isinstance(text, str):
        if visual_width(text) <= width and not order:
            return text

    diffed = False
    for i, word in enumerate(styled_words):
        used = line_len + visual_width(word)
        if used + margin > width:
            first_line = False
            if from_center:
                text = " ".join(styled_words[pos:i])
                centered += align.center(text.strip(), 
                    line=l, hue=h, line_hue=lh) + "\n"
                pos = i
                res = word
            else: result += '\n' + ' ' * indent + word
            line_len = indent + visual_width(word)
            if not diffed:
                width -= sub_indent
                diffed = True
        else:
            if iswhitespace(result): result += '' + word
            else: result += (' ' if result else '') + word
            if from_center: res += ' ' + word
            line_len += visual_width(word) + margin
    
    if _internal:
        rest = wrap_text(rest, indent, pad, inline, 
               order, from_center, _internal)
        result += f"\n{rest}" if rest else rest
    if not from_center: return result
    return centered + align.center(res, l, h, lh)

def pathit(path_str: str, length: int = 3) -> str:
    """Abbreviates a path if it becomes too long."""
    if not path_str: return path_str
    path_str = str(path_str)
    prefix   = f"{os.sep}...{os.sep}"
    cwd      = path_str
    if len(cwd.split(os.sep)) < length: return cwd
    return prefix + os.sep.join(cwd.split(os.sep)[-length:])

def has_unicode(s:str) -> bool:
    strip_ansi(s)
    for ch in s:
        if isunicode(ch): return True
    return False

def isunicode(s:str) -> bool:
    strip_ansi(s)
    return ord(s) in [9, 10, 13, 27]

def preserve_codes(s:str) -> str:
    strip_ansi(s)
    codes = {
          "\n": "\\n",
          "\r": "\\r",
          "\t": "\\t",
        "\x1b": "\\x1b",
        "\033": "\\033"
    }
        
    preserved = [ch for ch in s]
    for c, r in codes.items():
        for i, ch in enumerate(preserved):
            if ch == c: preserved[i] = r
                
    return "".join(preserved)

def label(iterable: list|tuple|dict, hue:str = "cyan",
          ordered: bool = False) -> list[str]:
    if ordered: return [color(f"{i+1}. {head}", hue) 
                for i, head in enumerate(iterable)]
    return [style_text(head, fg=hue) for head in iterable]

def iter_print(text, times: int, end: str = "\n", 
               delay: int|float = 0):
    
    validate(
        [times, int, "positive integer", "eqless", 0],
        [end, str, "string"],
        [delay, [int, float], "natural number", "less", 0],
        err=err)
    
    for _ in range(times):
        time.sleep(delay)
        print(text, end=end, flush=True)

def pad_args(*args) -> list[str]:
    return [str(n).zfill(2) for n in args]
    
def err(cause:Any, required: str) -> NoReturn:
    raise InputError(cause=cause, required=required)
 


"""Custom exceptions module"""
from collections.abc import Iterable
from datetime import datetime as dt
from math import floor, ceil, sqrt
from typing import NoReturn, Any
from . import logictools
import unicodedata
import textwrap
import time
import re

def _notify(msg:str, fg: str|None, bg: str|None, bold:bool, 
            underline:bool, inline:bool, spaced:bool, 
            centered:bool, write: str|bool, 
            _internal:bool) -> str|None|NoReturn:
    """private helper for non-flow-breaking warnings"""
    msg = color(msg, fg, bg, bold, underline)
    centered_msg = color(center(msg), fg, bg, bold)
    if inline: return msg
    if not centered:
        if spaced: msg = f"\n{msg}\n"
    else:
        if not spaced: msg = centered_msg
        else: msg = f"\n{centered_msg}\n"
    
    if not write: print(msg)
    
    if isinstance(write, str) and path.isfile(write):
        with open(write, "a+") as file:
            file.write(f"{msg}\n\n")
    elif write and not _internal: 
        warning("Warning! File not found", _internal=True)
    
    return msg
    
def empty(msg:str = "No contents found",
          bg: str|None = None, bold:bool = False,
          inline:bool = False, spaced:bool = True, 
          underline:bool = False, centered:bool = True, 
          write: str|bool = False, 
          _internal:bool = False) -> str | None:
    """Prints a gray colored warning message"""
    return _notify(msg, "gray", bg, bold, underline,
                   inline, spaced, centered, write,
                   _internal)

def warning(msg:str = "Warning!", bg: str|None = None, 
            bold:bool = False, inline:bool = False, 
            spaced:bool = True, underline:bool = False, 
            centered:bool=True, write: str|bool = False,
            _internal:bool = False) -> str | None:
    """Prints a yellow colored warning message"""
    return _notify(msg, "yellow", bg, bold, underline,
                   inline, spaced, centered, write,
                   _internal)
    
def validate(*argtypes, err=None) -> bool|NoReturn:
    """
Input validator

Args:
    argtypes: normally a collect of lists with the inputs
              needing validation. Each list should be
              either like this:
              [input(s), type(s), expected], or
              [input(s), type(s), expected, condition, exp]
              If there are multiple inputs or types, they
              should be in a list.
              <expected> is a string of the input that was
              required.
              <condition>: a string representing comparison
                           operator. The strings are
                           'less' (<), 'eqless' (<=),
                           'equal' (==), 'eqgreat' (>=),
                           'greater (>). They are for
                           comparing inputs against
                           expected input.
              <exp>: Expected [starting/ending] argument.
                     e.g., input 'less' exp (input < exp)
                     checks that the input is less than
                     exp else it throws an error.
    """
    def compare(arg1, against, arg2) -> bool:
        if against ==    "less": return arg1  < arg2
        if against ==  "eqless": return arg1 <= arg2
        if against ==   "equal": return arg1 == arg2
        if against == "eqgreat": return arg1 >= arg2
        if against == "greater": return arg1  > arg2
    
    for argtype in argtypes:
        args  = argtype[0]
        types = flatten(argtype[1], to=tuple)
        req   = argtype[2]
        
        if len(argtype) < 4:
            if isinstance(args, (list, tuple)):
                for arg in args:
                    if not isinstance(arg, types):
                        err(arg, req)
                continue
            
            if not isinstance(args, types): err(args, req)
            continue
                    
        against = argtype[3]
        arg2    = argtype[4]
        if isinstance(args, (str, int, float)):
            if not isinstance(args, types) or compare(
                args, against, arg2): err(args, req)
            continue
        
        for arg in args:
            if not isinstance(arg, types) or compare(
                arg, against, arg2): err(arg, req)
    
def ExitError() -> tuple:
    """function for catching exit-type errors"""
    return (EOFError, KeyboardInterrupt,)

class CUIError(Exception):
    """Base class for all custom CUI exceptions."""
    def __init__(self, msg: str = "An error occurred", 
                 hue: str | None = "red"):
        super().__init__(color(msg, hue))
        self.msg = msg
        self.now = dt.now()

    def display(self, inline: bool = False, 
                centered:bool = False,fg: str|None = "red", 
                bg: str | None = None, bold: bool = False, 
                underline: bool = False) -> None:
        """
Prints the exception message in red, optionally inline, centered, or underlined
        """
        if centered: self.msg = center(self.msg)
        msg = color(self.msg, fg, bg, bold, underline)
        print(msg, end='' if inline else '\n')
    
    @staticmethod
    def mod_error_tag(error:str) -> str:
        return f"tuikit.exceptions.{error}: "
    
    @staticmethod
    def cause_req(msg:str, cause:Any, required:str) -> str:
        def resolve_cause(c:Any) -> str:
            if type(c).__name__ == "function":
                length, c = len(c.__name__), c.__name__
                if length < 8: c = c + "(...)"
                else: c = c[:4] + "...(...)"
            elif "class" in str(c): c = str(c)[8:-2]
            
            if len(str(c)) > 12: c = str(c)[:9]+"..."
            return c
        
        if not any_eq(cause, required, eq='hapana'):
            if cause != 'hapana':
                cuz = resolve_cause(cause)
                type_name = type(cause).__name__
                if type_name == "type": 
                    type_name += " or function"
                if has_unicode(cause): 
                    type_name = "unicode" if len(cuz
                            ) < 2 else f"unicoded str"
                    cuz = preserve_codes(cuz)
                    if len(cuz) > 12: cuz = cuz[:9]+"..."
        
            if not all_eq(cause, required, eq='hapana'):
                msg =(f"'{cuz}' (type: {type_name}) is "
                    + f"not a valid {required}")
            elif cause != 'hapana':
                msg =(f"'{cuz}' (type: {type_name}) is "
                    + "not a valid input")
            else: msg = f"Invalid {required}"
    
        return msg

class Exceptions:
    """Custom exception factory"""
    def __getattr__(self, name):
        class __DynamicError(CUIError):
            def __init__(self, msg:str=f"{name} occurred", 
                         hue: str|None = "red"):
                super().__init__(f"[{name}] {msg}", hue)
        return __DynamicError

class TimeError(CUIError):
    """Raised for time-related errors"""
    def __init__(self, msg: str = "Invalid time format", 
                hue:str|None = "red", cause:Any = 'hapana', 
                required: str = 'hapana'):
        name = self.mod_error_tag("TimeError")
        msg  = self.cause_req(msg, cause, required)
        msg  = wrap_text(msg, list_order=name, inline=True)
        super().__init__(msg, hue)

class ListError(CUIError):
    """Raised for list-related errors"""
    def __init__(self, msg: str = "Invalid list format", 
                hue:str|None = "red", cause:Any = 'hapana', 
                required: str = 'hapana'):
        name = self.mod_error_tag("ListError")
        msg  = self.cause_req(msg, cause, required)
        msg  = wrap_text(msg, list_order=name, inline=True)
        super().__init__(msg, hue)

class InputError(CUIError):
    """Raised for invalid user input"""
    def __init__(self, msg: str = "Invalid input", 
                hue:str|None = "red", cause:Any = 'hapana', 
                required: str = 'hapana'):
        name = self.mod_error_tag("InputError")
        msg  = self.cause_req(msg, cause, required)
        msg  = wrap_text(msg, list_order=name, inline=True)
        super().__init__(msg, hue)


# ——————————————————————《 HELPERS 》———————————————————————

# Helpers are dumped down here to avoid circular
# import errors
# I did not use a utils.py module for these helpers
# as they are core functions of other modules

def any_eq(*args, eq:Any = None) -> bool:
    for arg in args:
        if arg == eq: return True
    return False

def all_eq(*args, eq:Any = None) -> bool:
    for arg in args:
        if arg != eq: return False
    return True

def flatten(data:Iterable, to:type = list) -> list|tuple:
    """
Recursively flattens any iterable (except str/bytes) into a flat list.

Args:
    data (Iterable): Input iterable, possibly nested.
    to (list or tuple): what to return the flattened
                        data as

Returns:
    list/tuple: Flattened version of the iterable.
    data: If input is not an iterable.
    """
    if not isinstance(data, Iterable) or isinstance(
        data, (str, bytes)): return data

    flat = []
    for item in data:
        if isinstance(item, Iterable) and not (
           isinstance(item, (str, bytes))):
           flat.extend(flatten(item))
        else: flat.append(item)
    
    return to(flat)

def strip_ansi(s: str) -> str:
    if not isinstance(s, str): __err__(s, "string") 
    return re.sub(r'\x1B[@-_][0-?]*[ -/]*[@-~]','',s)

def visual_width(s: str) -> int:
    clean = strip_ansi(s)
    width = 0
    for ch in clean:
        width += 2 if unicodedata.east_asian_width(ch) in [
                 'F', 'W'] else 1
    return width

def preserve_codes(s: str) -> str:
    codes = {
          "\n": "\\n",
          "\r": "\\r",
          "\t": "\\t",
        "\x1b": "\\x1b",
        "\033": "\\033"
    }
        
    preserved = [ch for ch in s]
    for c, r in codes.items():
        for i, ch in enumerate(preserved):
            if ch == c: preserved[i] = r
                
    return "".join(preserved)

def has_unicode(s: str) -> bool:
    try:      
        for ch in s:
            if isunicode(ch): return True
    except TypeError: return False
    return False

def isunicode(s: str) -> bool:
    return ord(s) in [9, 10, 13, 27]

def underline() -> None:
    term_width = logictools.get_term_width()
    print(color("—"*term_width, "magenta")) 

def color(text, fg: str|None = None, bg: str|None = None, 
          bold:bool = False, underline:bool = False)-> str:
    colors = {
        "black": 30, "red": 31, "green": 32, 
        "yellow": 33, "blue": 34, "magenta": 35,
        "cyan": 36, "white": 37,
        "gray": 90, "lightred": 91, "lightgreen": 92,
        "lightyellow": 93, "purple": 94,
        "lightmagenta": 95, "lightcyan": 96
    }

    if fg and fg not in colors:
        raise InputError(cause=fg)
    
    styles = []
    
    if bold: styles.append("1")
    if underline: styles.append("4")
    if fg in colors: styles.append(str(colors[fg]))
    if bg in colors: styles.append(str(colors[bg]+10))

    if styles: 
        return f"\033[{';'.join(styles)}m{text}\033[0m"
    return text

def center(text, lined:bool = False, doubled:bool = False, 
           line_hue: str|None = None, hue: str|None = None
           ) -> str:
    term_width  = logictools.get_term_size(True)
    display_len = visual_width(text)
    total_pad   = max(term_width - display_len, 0)
    left_pad    = total_pad // 2    
    right_pad   = total_pad - left_pad
    
    if lined or doubled:
        line   = "—" if lined else "="
        left   = color(line *  left_pad, fg=line_hue)
        right  = color(line * right_pad, fg=line_hue)
        middle = color(text, fg=hue)
        return f"{left}{middle}{right}"
    return " " * left_pad + color(text, fg=hue)

def wrap_text(text:str, indent:int = 0, pad:int = 0, 
              inline:bool = False, list_order:str = ''
              ) -> str:
    width = logictools.get_term_size()
    styled_words = text.split()
    
    length   = len(list_order) if inline else 0
    trailing = length / (10 ** len(str(length))
               ) if not inline else 0
    edge = 1 + trailing
    if length > 9 and not inline:
        trailing = length * 2 / 10
    margin = 1 + trailing if length > 9 else edge
    
    line_len = len(list_order)
    result   = list_order if not inline else ""
    if pad:
        result   = " " * (pad-1)
        line_len = pad
    for i, word in enumerate(styled_words):
        used = line_len + visual_width(word)
        if used + margin > width:
            result += '\n' + ' ' * indent + word
            line_len = indent + visual_width(word)
        else:
            result += (' ' if result else '') + word
            line_len += visual_width(
                word) + margin / (2 if margin >= 2 
                else 1)
    
    return result


from datetime import datetime, timedelta, tzinfo
from .exceptions import TimeError, InputError
from .__storage__ import CONTS, FILES, load
from .listools import flatten, list_items
from .console import underline
from .textools import Align

class Timezone():
    def __init__(self, zone: str|int|float = "CAT"):
        """
Initializes a Timezone object with the specified timezone

Args:
    zone: The timezone identifier (default is "CAT").
          Can be an IANA timezone (e.g., "Africa/
          Harare"), an abbreviation (e.g., "CAT") or 
          UTC offset
        

Raises:
    TimeError: If the timezone is not recognized
        """
        self.offset = self.get_offset(zone)
        self.zone = self.ZoneInfo(self.offset)
        self.name = self.zone.tzname(datetime.now())
        self.center = Align().center
        if isinstance(zone, str): self.name = zone

    class ZoneInfo(tzinfo):
        def __init__(self, offset):
            self.offset = timedelta(hours=offset)

        def utcoffset(self, dt): return self.offset

        def dst(self, dt): return timedelta(0)

        def tzname(self, dt):
            total_min = self.offset.total_seconds()/60
            hrs = int(total_min // 60)
            mins = int(abs(total_min) % 60)
            pn = "+" if hrs >= 0 else "-"
            return f"UTC{pn}{abs(hrs):02d}:{mins:02d}"

    def get_offset(self, tz:str|int|float)->int|float:
        def error():
            raise TimeError(cause=tz, required=
                "IANA timezone (e.g., 'Africa/"
                +"Harare'), abbreviation (e.g., 'CAT'"
                +") or UTC offset (e.g., 'UTC+2')")       
        
        if isinstance(tz, (int, float)): 
            if abs(tz) < 24: return tz
            error()
        try: 
            if "/" not in tz:
                tz = load("aliases.json")[tz.upper()]
        except (KeyError, TypeError):
            if tz.startswith("UTC"):
                try: 
                    ofs = float(tz[3:])
                    if abs(ofs) < 24: return ofs
                except ValueError: pass
            error()            
            
        info = tz.split("/")  
        tz = '/'.join(info)
        cont = info[0]
        for tz_cont, file in zip(CONTS, FILES):
            if cont == tz_cont:
                try: return load(file)[tz]
                except KeyError: break
        error()
                                
    def list_zones(self, aliased: bool = False,
              sort: str = "", getter: bool = False):
        """
Lists or returns available timezones or aliases

Args:
    aliased (bool): If True, show alias keys (e.g., 
                    'CAT')
    getter (bool): If True, return list instead of 
                   printing
    sort (str): filter list by continent or 
                abbreviation

Returns:
    list[str] (if getter=True): Sorted list of zones/
                                aliases
        """
        def error():
            raise InputError(cause=sort, required=
                f"filter for {chosen}")
        
        timezones = flatten([
            load(file).keys() for file in FILES if not 
            file.startswith("a")
        ])
        aliases = list(load("aliases.json").keys())
        
        chosen = "aliases" if aliased else "timezones"
        
        try: sort = sort.upper() if aliased else sort
        except AttributeError: error()
        
        temp = aliases if aliased else timezones        
        filt = [s for s in temp if s.startswith(sort)]
        if not filt: error()        
        if getter: return sorted(filt)
        
        header = self.center(f"《 {chosen.upper()} 》", 
                 "—", "magenta", "green")
        
        print(f"\n{header}\n\n")
        list_items(sorted(filt))
        print()
        underline(hue="magenta")
    
    @property
    def now(self) -> datetime:
        """Returns the current datetime in the configured zone"""
        return datetime.now(self.zone)

    def localize(self,*args)->datetime|list[datetime]:
        """
Converts naive or string datetime(s) to this timezone
Supports:
    - A single datetime or ISO string
    - Multiple datetime/ISO arguments
    - A list or tuple of datetime/ISO values
        """
        def error(dt="hapana"):
            if dt != "hapana":
                raise TimeError(cause=dt, required=
                "iso string (YYYY-MM-DD) or datetime"
                +" object")
            raise InputError("At least one datetime"
                +" or ISO string is required")
        
        def local(dt):
            if isinstance(dt, str):
                try: dt = datetime.fromisoformat(dt)
                except ValueError: error(dt)
            elif not isinstance(dt,datetime):error(dt)
            return dt.replace(tzinfo=self.zone) if (
                dt.tzinfo is None) else dt.astimezone(
                self.zone)

        # Handle list/tuple input
        if len(args)==1 and isinstance(args[0], (list, 
          tuple)):return [local(dt) for dt in args[0]]

        # Handle multiple arguments
        elif len(args) > 1:
            return [local(dt) for dt in args]

        # Handle single argument
        elif len(args) == 1: return local(args[0])

        error()

    def convert(self, dt:datetime, to:str)->datetime:
        """Converts a datetime object to a different timezone"""
        try: return dt.astimezone(Timezone(to).zone)
        except AttributeError: 
            req = "datetime object"
            raise TimeError(cause=dt, required=req)

    @property
    def iso_now(self) -> str:
        """Returns ISO string of the current time in this zone"""
        return self.now.isoformat()
    
    def diff_seconds(self, dt1, dt2) -> float:
        """
Returns total seconds difference between two datetimes
Both will be localized to self.zone before subtraction
        """
        dt1, dt2 = self.localize(dt1, dt2)
        return (dt2 - dt1).total_seconds()

if __name__ == "__main__":
    tz = Timezone()
    print(tz.iso_now)
    print(tz.localize(datetime.now().isoformat()))
    print(datetime.now().isoformat())
    tz.list_zones(sort="Africa")



import ast


MARKER = None


def tolerant_parse_module(source: list[str]|str,
                          get: bool = False) -> ast.Module:
    if isinstance(source, str): source = source.splitlines()
    nodes   = []
    blocks  = split_top_level_blocks(source)
    synerrs = []

    for block, lineno_offset in blocks:
        try:
            parsed = ast.parse('\n'.join(block))
            for node in parsed.body:
                shift_node_lines(node, lineno_offset - 1)
                nodes.append(node)
        except SyntaxError as e:
            lineno = e.lineno + lineno_offset - 1
            synerrs.append((lineno, e.msg))
            continue

    tree = ast.Module(body=nodes, type_ignores=[])
    annotate_parents(tree)
    if get: return tree, synerrs
    return tree


def annotate_parents(node, parent=None):
    if parent is None:
        parent = ast.Module(body=[], type_ignores=[])
    for child in ast.iter_child_nodes(node):
        child.parent = parent
        annotate_parents(child, child)


def in_docstring(marker):
    global MARKER
    marker = marker.strip()
    if MARKER and MARKER == marker: return False
    if MARKER and MARKER != marker: return True
    if not MARKER and marker != '"""': return False
    if not MARKER and marker == '"""':
        MARKER = marker
        return True


def line_indent(line: str) -> int:
    return len(line) - len(line.lstrip())


def is_top_level_start(line: str) -> bool:
    blocks = ("def ", "class ", "if ", "for ", "while ",
              "try", "@")
    if line.lstrip().startswith(blocks): return True
    elif line.lstrip().startswith("#"):
        return any(b in line for b in blocks)
    return False


def split_top_level_blocks(lines: list[str]
                          ) -> list[tuple[list[str], int]]:
    blocks       = []
    buffer       = []
    in_block     = False
    block_indent = 0
    start_lineno = 0
    for idx, line in enumerate(lines):
        if is_top_level_start(line) and line_indent(line) == 0:
            if buffer: blocks.append((buffer, start_lineno))
            buffer       = [line]
            in_block     = True
            block_indent = line_indent(line)
            start_lineno = idx + 1
        elif in_block:
            if line.strip() == "" or in_docstring(line) or line_indent(line) > block_indent:
                buffer.append(line)
            else:
                blocks.append((buffer, start_lineno))
                buffer       = [line]
                in_block     = is_top_level_start(line)
                start_lineno = idx + 1
        else:
            if not buffer: start_lineno = idx + 1
            buffer.append(line)

    if buffer: blocks.append((buffer, start_lineno))
    return blocks


def shift_node_lines(node, offset: int):
    for subnode in ast.walk(node):
        if hasattr(subnode, 'lineno'):
            subnode.lineno += offset
        if hasattr(subnode, 'end_lineno'):
            subnode.end_lineno += offset



from pathlib import Path
import json
import os

ZONES = Path(__file__).resolve().parent/"tzdata"
CONTS = [cont[:-5] for cont in os.listdir(ZONES)]
FILES = [     file for file in os.listdir(ZONES)]

def load(file):
    file_path = ZONES / file
    with open(file_path, "r") as f: return json.load(f)